# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.
import logging as log
import json
import sys
import os
import fcntl
import struct

from optparse import OptionParser
from subprocess import CalledProcessError
from collections import defaultdict
from contextlib import contextmanager

from plesk_file_utils import discard_stdout, redirect_stdout
import pum_errors


def _list_packages(package_manager, output_format, updates_only=False, exclude=None, repo_info=False):
    """ List of installed packages with updates info if available.
        Technically it simulates update transaction to detect
        which packages can be safely updated and extends the list with installed
        packages and "locked" fields if necessary.
    """
    if exclude is None:
        exclude = []
    with discard_stdout():
        transaction_info = package_manager.simulate_update(None)

    data = []
    pkg_names = set()

    for package in transaction_info.values():
        name = package["name"]
        pkg_names.add(name)
        item = {
            "name": name,
            "version_from": package["from"],
            "version_to": package["full_version"],
            "summary": package["summary"],
        }
        if repo_info:
            item["repo_from"] = package["repo_from"]
            item["repo_to"] = package["repo_to"]
        if name in exclude:
            item["locked"] = True
        data.append(item)

    if not updates_only:
        with discard_stdout():
            installed_packages = package_manager.get_installed_packages()
        for package in installed_packages.values():
            name = package["name"]
            if name in pkg_names:
                continue
            item = {
                "name": name,
                "version_from": package["full_version"],
                "summary": package["summary"],
            }
            if repo_info:
                item["repo_from"] = package["repo_from"]
            if name in exclude:
                item["locked"] = True
            data.append(item)

    data = sorted(data, key=lambda e: e["name"])

    if output_format == "json":
        json.dump(data, sys.stdout)
        sys.stdout.write("\n")
        return

    for package in data:
        package_info = "Package '%s' version '%s' " % (package["name"], package["version_from"])

        available_update = package.get("version_to")
        if available_update:
            update_info = "updatable to version '%s'" % (available_update)
            if repo_info:
                update_info += " (repo: '%s' -> '%s')" % (package["repo_from"], package["repo_to"])
        else:
            update_info = "up-to-date"
            if repo_info:
                update_info += " (repo: '%s')" % package["repo_from"]

        print package_info, "is", update_info


def _print_package_details(package_manager, package):
    """ Print package information using system package manager
        via commands "dpkg -s" or "rpm -qi"
    """
    package_manager.print_package_details(package)


def _simulate_update(package_manager, packages, output_format, exclude=None, repo_info=False):
    """ Simulate update of specified packages (all packages if None)
        and show transaction information.
    """
    if exclude is None:
        exclude = []

    with discard_stdout():
        package_manager.exclude_packages(exclude)
        transaction_info = package_manager.simulate_update(packages)

    data = []
    for package in transaction_info.values():
        item = {
            "name": package["name"],
            "version_from": package["from"],
            "version_to": package["full_version"],
            "summary": package["summary"],
        }
        if repo_info:
            item["repo_from"] = package["repo_from"]
            item["repo_to"] = package["repo_to"]
        data.append(item)

    data = sorted(data, key=lambda e: e["name"])

    if output_format == "json":
        json.dump(data, sys.stdout)
        sys.stdout.write("\n")
        return

    for p in data:
        info = "Update %s from %s to %s" % (p["name"], p["version_from"], p["version_to"])
        if repo_info:
            info += " (repo: '%s' -> '%s')" % (p["repo_from"], p["repo_to"])
        print(info)


def _update_packages(package_manager, packages, exclude=None):
    """ Safely update specified packages (or all packages if None)
        respecting exclude list.
    """
    if exclude is None:
        exclude = []

    with redirect_stdout(sys.stderr):
        package_manager.exclude_packages(exclude)
        package_manager.update_packages(packages)


def _check(package_manager, full):
    """ Check consistency of package manager and packages state. """
    package_manager.check(full)


def _list_compatible_repos(package_manager):
    dict_compat = defaultdict(list)
    for (r1, r2, compat) in package_manager.get_compatible_repos_list():
        if compat:
            dict_compat[r1].append(r2)
    keys = sorted(dict_compat.keys())
    for r in keys:
        repo_list = "'" + ("', '".join([str(x) for x in dict_compat[r]])) + "'"
        print "Packages from '%s' can be updated from repositories: %s" % (r, repo_list)


def _get_logging_verbosity(verbose):
    default_level = 3  # INFO
    return default_level + verbose


def _configure_logging(verbosity):
    log_level = log.INFO
    if verbosity <= 0:
        log_level = log.CRITICAL
    elif verbosity == 1:
        log_level = log.ERROR
    elif verbosity == 2:
        log_level = log.WARNING
    elif verbosity == 3:
        log_level = log.INFO
    elif verbosity >= 4:
        log_level = log.DEBUG

    log.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s",
        level=log_level,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    return log_level


def _check_for_panel_upgrade_errors():
    if ((os.path.exists('/var/lock/parallels-panel-maintenance-mode.flag') or
         os.path.exists('/var/lock/parallels-panel-upgrade-failure.flag'))):
        raise pum_errors.PanelUpgradeError()


@contextmanager
def _locking_autoinstaller():
    if not sys.platform.startswith("linux"):
        raise Exception("Only 'linux*' platforms are supported (found: %s)" % sys.platform)

    l_whence = os.SEEK_SET
    l_start = 0
    l_len = 1

    with os.fdopen(os.open("/tmp/psa-installer.lock", os.O_WRONLY | os.O_CREAT, 0600), "w") as lockfile:
        try:
            flock = struct.pack("hhllhh", fcntl.F_WRLCK, l_whence, l_start, l_len, 0, 0)
            fcntl.fcntl(lockfile.fileno(), fcntl.F_SETLK, flock)
        except IOError:
            raise pum_errors.AILockError()

        try:
            _check_for_panel_upgrade_errors()

            yield
        finally:
            funlock = struct.pack("hhllhh", fcntl.F_UNLCK, l_whence, l_start, l_len, 0, 0)
            fcntl.fcntl(lockfile.fileno(), fcntl.F_SETLK, funlock)


def _do_main(package_manager_maker):
    OPTIONS = [
        # short long            action          default     help
        ("-l",  "--list",       "store_true",   False,      "list installed packages"),
        ("-i",  "--info",       "store_true",   False,      "show information on the packages"),
        ("-s",  "--simulate",   "store_true",   False,      "simulate an update of the packages"),
        ("-u",  "--update",     "store_true",   False,      "update the packages"),
        ("-c",  "--check",      "store_true",   False,      "check consistency of package manager state"),
        ("-j",  "--json",       "store_true",   False,      "output using JSON"),
        ("-U",  "--updates-only", "store_true", False,      "list only packages with available updates"),
        ("-R",  "--repo-info",  "store_true",   False,      "Show repository sources for packages"),
        ("-C",  "--list-compatible-repos",  "store_true",   False,      "List repositories with compatibility info"),

        ("-x",  "--exclude",    "append",       [],         "exclude packages from update"),
        ("-f",  "--full",       "store_true",   False,      "perform long running checks"),
        ("-S",  "--unsafe",     "store_true",   False,      "unsafe mode"),
        ("-v",  "--verbose",    "count",        0,          "increase verbosity"),
    ]

    option_parser = OptionParser(usage="pum [options] [packages]")
    for option in OPTIONS:
        option_parser.add_option(option[0], option[1], action=option[2], default=option[3], help=option[4])

    (options, arguments) = option_parser.parse_args()

    verbosity = _get_logging_verbosity(options.verbose)
    log_level = _configure_logging(verbosity)

    log.info("pum is called with arguments: %s" % sys.argv[1:])
    if log_level == log.DEBUG:
        log.debug("pum is started with debug logging")

    if options.list + options.info + options.simulate + options.update + options.check > 1:
        option_parser.error("options -l, -i, -s, -u and -c are mutually exclusive")

    with _locking_autoinstaller():
        _do_work(package_manager_maker, options, arguments)

    log.info("pum completed successfully")


def _do_work(package_manager_maker, options, arguments):
    output_format = "json" if options.json else "unspecified"

    with discard_stdout():
        package_manager = package_manager_maker(options.unsafe)

    if options.list:
        with package_manager.locking():
            _list_packages(package_manager, output_format, options.updates_only, options.exclude, options.repo_info)
    elif options.info:
        with package_manager.locking():
            _print_package_details(package_manager, arguments[0])
    elif options.simulate:
        with package_manager.locking():
            _simulate_update(package_manager, arguments, output_format, options.exclude, options.repo_info)
    elif options.update:
        with package_manager.locking():
            _update_packages(package_manager, arguments, options.exclude)
    elif options.check:
        # do not lock here: locking yumdb from pum leads to deadlock
        _check(package_manager, options.full)
    elif options.list_compatible_repos:
        with package_manager.locking():
            _list_compatible_repos(package_manager)
    else:
        raise Exception("invalid usage")


def do_main(package_manager_maker):
    exitcode = 0
    try:
        _do_main(package_manager_maker)
    except CalledProcessError as ex:
        exitcode = ex.returncode
    except pum_errors.AILockError:
        log.error("Update operation was locked by another update process (Plesk installer or pum).")
        exitcode = pum_errors.LOCK_ERROR_EXIT_CODE
    except pum_errors.PMLockError:
        log.error("System package manager is already running.")
        exitcode = pum_errors.LOCK_ERROR_EXIT_CODE
    except pum_errors.PanelUpgradeError:
        log.error("Previous Plesk installer call finished unsuccessfully. PUM will not perform any update operations until Plesk installation is fixed.")
        exitcode = pum_errors.PANEL_UPGRADE_ERROR_EXIT_CODE
    except SystemExit as ex:
        exitcode = ex.code
    except Exception as ex:
        log.error(ex)
        log.debug(ex, exc_info=True)
        exitcode = 1
    finally:
        if exitcode != 0:
            log.error("Exited with returncode %d.", exitcode)
        sys.exit(exitcode)

# vim: ts=4 sts=4 sw=4 et :
