# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.
from optparse import OptionParser
import sys
import os
import logging
import urllib2
from contextlib import closing

import plesk_subprocess as subprocess
from plesk_constants import PRODUCT_ROOT_D


def _get_ext_name():
    try:
        return os.environ["PLESK_EXTENSION"]
    except KeyError:
        raise Exception("This utility should be called inside extension installation script")


class AptRepoHelper:
    def create(self, opts):
        if not opts.source:
            raise Exception("Source string is unspecified")
        path = self._get_repo_config_path(opts.temporary)
        file_mode = "a" if opts.append else "w"
        with open(path, file_mode) as f:
            print >> f, opts.source

    def delete(self, opts):
        try:
            os.remove(self._get_repo_config_path(opts.temporary))
        except OSError:
            pass

    def import_gpg_key(self, opts):
        if opts.gpg_key:
            with closing(urllib2.urlopen(opts.gpg_key)) as f:
                key = f.read()
            p = subprocess.Popen(["apt-key", "add", "-"], stdin=subprocess.PIPE)
            p.communicate(input=key)
            if p.returncode != 0:
                raise Exception("Failed to import gpg key: 'apt-key add' finished with exit code %d" % p.returncode)
        if opts.gpg_fingerprint:
            if not opts.gpg_key_servers:
                raise Exception("Key servers should be specified for importing gpg-key by fingerprint")
            for server in opts.gpg_key_servers.split():
                try:
                    subprocess.check_call(["apt-key", "adv", "--keyserver", ("hkp://%s:80" % server), "--recv-keys", opts.gpg_fingerprint])
                    return
                except:
                    pass
            raise Exception("gpg key (%s) is not imported" % opts.gpg_fingerprint)

    def _get_repo_config_path(self, temporary):
        ext_name = _get_ext_name()
        return "/etc/apt/sources.list.d/plesk-ext-%s%s.list" % (ext_name, (".temp" if temporary else ""))


class YumRepoHelper:
    def create(self, opts):
        if not opts.baseurl and not opts.mirror_list:
            raise Exception("Neither baseurl nor mirror list is specified")
        path = self._get_repo_config_path(opts.temporary)
        file_mode = "a" if opts.append else "w"
        with open(path, file_mode) as f:
            ext_name = _get_ext_name()
            if opts.id:
                repo_id = opts.id
            else:
                repo_id = "plesk-ext-" + ext_name
                if opts.temporary:
                    repo_id += "-temp"
            print >> f, "[%s]" % repo_id
            repo_desc = opts.description if opts.description else ("%s extension repository" % ext_name)
            print >> f, "name=%s" % repo_desc
            if opts.baseurl:
                print >> f, "baseurl=%s" % opts.baseurl
            elif opts.mirror_list:
                print >> f, "mirrorlist=%s" % opts.mirror_list
            print >> f, "enabled=1"
            if opts.gpg_key:
                if opts.no_gpg_check:
                    raise Exception("Both gpg-key and no-gpg-check are specified")
                print >> f, "gpgcheck=1\ngpgkey=%s" % opts.gpg_key
            elif opts.no_gpg_check:
                print >> f, "gpgcheck=0"

    def delete(self, opts):
        try:
            os.remove(self._get_repo_config_path(opts.temporary))
        except OSError:
            pass

    def import_gpg_key(self, opts):
        if opts.gpg_key:
            if opts.create:
                # gpg-key will be added for repository specifically
                pass
            else:
                raise Exception("Global import of gpg-key for yum is not supported. Use per-repository gpg-settings instead.")

        if opts.gpg_fingerprint:
            raise Exception("Importing gpg-key by fingerprint for yum is not supported")

    def _get_repo_config_path(self, temporary):
        ext_name = _get_ext_name()
        return "/etc/yum.repos.d/plesk-ext-%s%s.repo" % (ext_name, (".temp" if temporary else ""))


def get_pkg_type():
    cmd = [os.path.join(PRODUCT_ROOT_D, "admin/sbin/osdetect"), "--pkg"]
    return subprocess.check_output(cmd).strip()


def main():
    try:
        parser = OptionParser()
        parser.add_option("--create", action="store_true", default=False, help="create repository configuration")
        parser.add_option("--delete", action="store_true", default=False, help="delete repository configuration")
        parser.add_option("--source", action="store", help="[apt] source string for repository")
        parser.add_option("--baseurl", action="store", help="[yum] repository url")
        parser.add_option("--mirror-list", action="store", help="[yum] mirrorlist url")
        parser.add_option("--id", action="store", help="[yum] override default generated repository id (good to use with --append)")
        parser.add_option("--description", action="store", help="[yum] override default generated repository description")
        parser.add_option("--persistent", action="store_true", default=False, help="create persistent repository")
        parser.add_option("--temporary", action="store_true", default=False, help="create temporary repository")
        parser.add_option("--append", action="store_true", default=False, help="append record to existing repo config")
        parser.add_option("--gpg-key", action="store", help="url to repository gpg key")
        parser.add_option("--gpg-fingerprint", action="store", help="[apt] gpg key fingerprint")
        parser.add_option("--gpg-key-servers", action="store", help="[apt] space separated list of gpg key servers")
        parser.add_option("--no-gpg-check", action="store_true", default=False, help="[yum] skip gpg check for repository")
        (opts, args) = parser.parse_args()

        if not (opts.create or opts.delete or opts.gpg_key or opts.gpg_fingerprint):
            raise Exception("Some of --create, --delete, --gpg-key, --gpg-fingerprint should be specified")

        if opts.create and opts.delete:
            raise Exception("Only one of --create, --delete should be specified")

        if (opts.create or opts.delete) and len(filter(None, [opts.persistent, opts.temporary])) != 1:
            raise Exception("One of --persistent, --temporary should be specified")

        pkg = get_pkg_type()
        if pkg == "deb":
            helper = AptRepoHelper()
        elif pkg == "rpm":
            helper = YumRepoHelper()
        else:
            raise Exception("Unknown package type: '%s'" % pkg)

        if opts.gpg_key or opts.gpg_fingerprint:
            helper.import_gpg_key(opts)

        if opts.create:
            helper.create(opts)
        elif opts.delete:
            helper.delete(opts)

    except Exception as e:
        logging.error(e)
        logging.debug(e, exc_info=True)
        sys.exit(1)


if __name__ == "__main__":
    main()
