# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.

import sys
import os
import json

from optparse import OptionParser
import errno
import plesk_subprocess as subprocess
import plesk_log
from plesk_constants import PRODUCT_ROOT_D

USAGE = '''
  %prog [options] <command> [command-arguments]

Commands:
  versions              display the versions of Ruby interpreters known to
                        Plesk
  info <version>        display info on the specified version of Ruby
                        interpreter
  register <version> <path-to-ruby-root>
                        make the specified version of Ruby interpreter
                        known to Plesk
  unregister <version>  make the specified version of Ruby interpreter
                        unknown to Plesk
  get-version <user> <directory>
                        returns ruby version from specified directory of sysuser
  set-version <user> <directory> <version>
                        set specified version to specified directory of sysuser
  run-bundler <user> <directory>
                        run bundler install as user
'''

RUBY_VERSIONS_DIR = '/etc/psa/ruby_versions.d/'
RBENV_VERSIONS_DIR = '/var/lib/rbenv/versions/'
RUBY_VERSIONS_LIST = '/var/lib/plesk/ruby_versions.json'

log = plesk_log.getLogger(__name__)


class Ruby(object):

    def __init__(self, **kwargs):
        for keyword in ["version", "path"]:
            setattr(self, keyword, kwargs[keyword])
        self.path_to_ruby = os.path.join(self.path, "bin/ruby")
        try:
            self.fullversion = subprocess.check_output(
                [self.path_to_ruby, '-e', 'print "#{RUBY_VERSION}-p#{RUBY_PATCHLEVEL}"'])
        except Exception as e:
            raise type(e)('cannot get version for %s (%s): %s' %
                          (self.path, self.version, str(e)))

    def _to_string(self):
        return "%s %s" % (self.version, self.path)

    def _to_dict(self):
        output = {}
        for k in ['version', 'fullversion', 'path']:
            output[k] = getattr(self, k)
        return output


class RubyJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Ruby):
            obj = obj._to_dict()
        else:
            obj = super(RubyJsonEncoder, self).default(obj)
        return obj


def _get_permanent_versions():
    versions = {}

    if not os.path.isdir(RUBY_VERSIONS_DIR):
        return versions

    for entry in os.listdir(RUBY_VERSIONS_DIR):
        filename = os.path.join(RUBY_VERSIONS_DIR, entry)
        if not os.path.isfile(filename):
            continue

        try:
            with open(filename, 'r') as descriptor:
                version_info = json.load(descriptor)
                r = Ruby(**version_info)

                if r.version in versions:
                    log.warn('duplicate version `' + r.version + '`')

                versions[r.version] = r
        except Exception as e:
            log.warn('could not process file ' + filename + ': ' + str(e))

    return versions


def _get_custom_versions():
    versions = {}

    if os.path.isfile(RUBY_VERSIONS_LIST):
        try:
            with open(RUBY_VERSIONS_LIST, 'r') as input_file:
                for version, version_info in json.load(input_file).iteritems():
                    try:
                        r = Ruby(**version_info)
                    except Exception as e:
                        log.warning("Bad ruby version: %s" % str(e))
                        continue
                    versions[version] = r
        except Exception as e:
            log.warn('could not load custom versions file ' +
                     RUBY_VERSIONS_LIST + ': ' + str(e))
    return versions


def _set_custom_versions(versions):
    with open(RUBY_VERSIONS_LIST, 'w') as output_file:
        json.dump(versions, output_file, cls=RubyJsonEncoder)


def _get_known_versions():
    versions = _get_custom_versions()
    try:
        versions.update(_get_permanent_versions())
    except OSError:
        pass
    return versions


def _show_versions(output_format):
    versions = _get_known_versions()

    if output_format == 'json':
        json.dump(versions, sys.stdout, cls=RubyJsonEncoder)
        return

    if output_format == 'plain':
        for v in versions.values():
            print v._to_string()
        return


def _show_info(output_format, version):
    versions = _get_known_versions()

    version_info = versions.get(version)
    if not version_info:
        return

    if output_format == 'json':
        json.dump(version_info, sys.stdout, cls=RubyJsonEncoder)
        return

    if output_format == 'plain':
        print version_info._to_string()
        return


def _set_version(version, path):
    if version in _get_permanent_versions():
        raise Exception(
            'the specified version has been registered by a package; it can be updated with editing the corresponding file in ' + RUBY_VERSIONS_DIR)

    versions = _get_custom_versions()

    version_info = versions.get(version)
    if not version_info:
        versions[version] = Ruby(version=version, path=path)
    else:
        version_info.path = path

    _set_custom_versions(versions)


def _discard_version(version):
    if version in _get_permanent_versions():
        raise Exception('the specified version has been registered by a package; it can be unregistered with either uninstalling the package or removing the corresponding file from ' + RUBY_VERSIONS_DIR)

    versions = _get_custom_versions()
    if version not in versions:
        return

    del versions[version]

    _set_custom_versions(versions)
    try:
        os.unlink(os.path.join(RBENV_VERSIONS_DIR, version))
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def _rehash_rbenv_versions():
    versions = _get_known_versions()

    for e in os.listdir(RBENV_VERSIONS_DIR):
        path = os.path.join(RBENV_VERSIONS_DIR, e)
        if os.path.islink(path) and not os.path.exists(path):
            log.debug('Removing stale link %s' % path)
            os.unlink(path)

    for version, info in versions.iteritems():
        if not os.path.exists(os.path.join(RBENV_VERSIONS_DIR, version)):
            os.symlink(info.path, os.path.join(RBENV_VERSIONS_DIR, version))


def _get_user_ruby_version(username, directory):
    return subprocess.check_output([
        os.path.join(PRODUCT_ROOT_D, 'admin/sbin/filemng'),
        username, 'exec', directory,
        '/bin/bash', '-l', '-c', 'rbenv version-name'
    ]).strip()


def _set_user_ruby_version(username, directory, version):
    return subprocess.check_call([
        os.path.join(PRODUCT_ROOT_D, 'admin/sbin/filemng'),
        username, 'exec', directory,
        '/bin/bash', '-l', '-c', 'rbenv local %s' % version
    ])


def _run_user_bundler(username, directory):
    return subprocess.check_call([
        os.path.join(PRODUCT_ROOT_D, 'admin/sbin/filemng'),
        username, 'exec', directory,
        '/bin/bash', '-l', '-c', 'bundle install --path vendor/bundle'
    ])


def doMain():
    plesk_log.setDefaultLevel('warning')

    option_parser = OptionParser(usage=USAGE)
    option_parser.description = 'An utility for managing versions of Ruby interpreters in Plesk'
    option_parser.add_option('-f', '--output-format', metavar='format',
                             type='choice', choices=['plain', 'json'], default='plain',
                             help='output info in the specified format (plain/json)')

    (options, arguments) = option_parser.parse_args()
    if not arguments:
        option_parser.error('missing command')

    command = arguments[0]
    arguments = arguments[1:]

    if command == 'versions':
        _show_versions(options.output_format)
    elif command == 'info':
        if len(arguments) != 1:
            option_parser.error('inappropriate number of command arguments')
        _show_info(options.output_format, arguments[0])
    elif command == 'register':
        if len(arguments) != 2:
            option_parser.error('inappropriate number of command arguments')
        _set_version(arguments[0], arguments[1])
    elif command == 'unregister':
        if len(arguments) != 1:
            option_parser.error('inappropriate number of command arguments')
        _discard_version(arguments[0])
    elif command == 'get-version':
        if len(arguments) != 2:
            option_parser.error('inappropriate number of command arguments')
        version = _get_user_ruby_version(arguments[0], arguments[1])
        if options.output_format == 'json':
            json.dump({'version': version}, sys.stdout)
        else:
            print version
    elif command == 'set-version':
        if len(arguments) != 3:
            option_parser.error('inappropriate number of command arguments')
        _set_user_ruby_version(arguments[0], arguments[1], arguments[2])
    elif command == 'run-bundler':
        if len(arguments) != 2:
            option_parser.error('inappropriate number of command arguments')
        _run_user_bundler(arguments[0], arguments[1])
    else:
        option_parser.error('command `' + command + '` not recognized')

    _rehash_rbenv_versions()


def main():
    try:
        doMain()
    except SystemExit:
        raise
    except Exception as e:
        log.error(e)
        log.debug(e, exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
