# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.

from optparse import OptionParser
import os
import sys
import logging as log
import plesk_subprocess as subprocess
import re
import modules as modules

from common import *

DHPARAMS_LEN = 2048
VERBOSE = 0

def condrestart(service):
    """ restart service if it started, or do nothing """
    try:
        PRODUCT_ROOT_D = "/opt/psa"
        pleskrc = os.path.join(PRODUCT_ROOT_D, 'admin/sbin/pleskrc')
        subprocess.check_call([pleskrc, service, "status"])
    except subprocess.CalledProcessError, e:
        log.debug("Service %s stopped (%d), no reload", service, e.returncode)
        return

    log.debug("reload service %s", service)
    subprocess.check_call([pleskrc, service, "reload"])

def normalize_ciphers(ciphers):
    return ciphers.strip()

def normalize_protocols(protocols_string):
    protocols = []
    exclude = []

    # Note, protocols it should be sorted by strength, as we need to check least sufficient protocol version for courier
    SUPPORTED_PROTOCOLS = "TLSv1 TLSv1.1 TLSv1.2".split()

    supported_protocols = dict(zip(map(str.lower, SUPPORTED_PROTOCOLS), SUPPORTED_PROTOCOLS))

    for p in re.split(r"\s+|:", protocols_string):
        try:
            if p.startswith('!'):
                exclude.append(supported_protocols[p[1:].lower()])
            else:
                protocols.append(supported_protocols[p.lower()])
        except KeyError:
            log.warning("Ignoring unsuppored protocol %s", p)

    if protocols and exclude:
        raise ValueError("Illegal mix of enable and disable(!) protocols in %s", protocols_string)

    if protocols:
        return protocols

    if exclude:
        return [x for x in SUPPORTED_PROTOCOLS if x not in exclude]

def get_ciphers():
    r = re.compile(r'^ssl-ciphers:\s(.*?)\s*$', re.MULTILINE)
    output = subprocess.check_output(['/usr/local/psa/bin/server_pref', '--show'])
    m = r.search(output)
    return m.group(1)

def get_protocols():
    r = re.compile(r'^ssl-protocols:\s(.*?)\s*$', re.MULTILINE)
    output = subprocess.check_output(['/usr/local/psa/bin/server_pref', '--show'])
    m = r.search(output)
    return m.group(1)

def flush_custom_options(service):
    cfg_path = "/var/lib/plesk/ssl_" + service + ".conf"
    try:
        os.unlink(cfg_path)
    except OSError, e:
        if e.errno == 2:
            return
        raise Exception("Unable to remove saved cache for '%s' service: %s (%i)" % (service, e.strerror, e.errno))

def show_customized(service):
    cfg_path = "/var/lib/plesk/ssl_" + service + ".conf"
    if not os.path.exists(cfg_path):
        print service.ljust(12) + ": default"
        return

    print service.ljust(12) + ": custom"
    ciphers, protocols = get_custom_options(service, "", "", verbose = 0)
    log.info("    protocols: " + protocols)
    log.info("    ciphers:   " + ciphers + "\n")

    return

def get_custom_options(service, ciphers, protocols, update_cache = False, verbose = 0):
    cfg_path = "/var/lib/plesk/ssl_" + service + ".conf"

    if not update_cache:
        res = {}
        try:
            with open(cfg_path, 'r') as fobj:
                if verbose >= 1:
                    print "Ciphers and protocols were not changed for '" + service + "' because of custom configuration is used."
                for line in fobj.readlines():
                    data = line.strip().split()
                    res[data[0]] = ' '.join(data[1:])
                fobj.close()
        except IOError, e:
            if e.errno == 2:
                return ciphers, ' '.join(protocols)
            raise Exception("Unable to read data from file '%s': %s (%i)" % (cfg_path, e.strerror, e.errno))

        return res["ciphers"], res["protocols"]

    proto = ' '.join(protocols)
    config_set("ciphers", "ciphers %s" % ciphers, cfg_path)
    config_set("protocols", "protocols %s" % proto, cfg_path)

    return ciphers, proto

def do_main():
    global DHPARAMS_LEN
    global VERBOSE
    parser = OptionParser(usage="Usage: %prog [-v]")
    parser.add_option("-v", "--verbose", action="count", help="Increase verbosity (can be specified multiple times).")
    parser.add_option("-c", "--ciphers", default=get_ciphers(), type="string", help="Set selected SSL ciphers.")
    parser.add_option("-p", "--protocols", default=get_protocols(), type="string", help="Set selected SSL protocols.")
    parser.add_option("-s", "--services", type="string", help="Configure only selected services.")

    parser.add_option("-l", "--show-custom", action="store_true", help="Show customized services list.")
    parser.add_option("-C", "--custom", action="store_true", default=False, help="Save per-service ciphers and protocols")
    parser.add_option("-f", "--no-custom", action="store_true", default=False, help="Reset custom per-service ciphers and protocols to default.")

    parser.add_option("-t", "--certificate", action="store_true", default=False, help="Do actions with SSL certificate for the service.")
    parser.add_option("-z", "--cert", type="string", default="", help="Specify the certificate file.")
    parser.add_option("-k", "--private-key", type="string", default="", help="Specify the private key file.")
    parser.add_option("-r", "--ca-chain", type="string", default="", help="Specify the file with intermediate and root CA certificates.")

    parser.add_option("-T", "--disable-tls-compression", action="store_true", default=False, help="Disable TLS compression ([CRIME]).")
    parser.add_option("-D", "--strong-dh", action="store_true", default=False, help="Secure Diffie-Hellman ([Logjam]).")
    parser.add_option("-G", "--regenerate-dhparams", action="store_true", default=False, help="Regenerate dhparams.pem.")
    parser.add_option("-S", "--dhparams-size", type="int", default=DHPARAMS_LEN, help="Specify dh parameters size (default: %d)." % DHPARAMS_LEN)

    (options, args) = parser.parse_args()

    log_level = log.WARNING
    VERBOSE = options.verbose
    if options.verbose == 1:
        log_level = log.INFO
        log.basicConfig(format='%(message)s', level=log_level)
    elif options.verbose >= 2:
        log_level = log.DEBUG

    log.basicConfig(format='%(levelname)s:%(message)s', level=log_level, stream=sys.stderr)

    if args:
        raise ValueError("No positional arguments are accepted by this utility. Got: %s" % args)

    ciphers = normalize_ciphers(options.ciphers)
    protocols = normalize_protocols(options.protocols)

    services = modules.SERVICES.keys()
    if options.services:
        services = options.services.split()

    if not protocols and not options.disable_tls_compression and not options.strong_dh and not options.certificate:
        raise ValueError("No supported protocols supplied")

    if not ciphers and not options.disable_tls_compression and not options.strong_dh and not options.certificate:
        raise ValueError("No supported ciphers supplied")

    if len(filter(None, [protocols, ciphers])) == 1:
        raise ValueError("Both ciphers and protocols should be specified")

    if options.custom and options.no_custom:
        raise ValueError("Both '--custom' and '--no-custom' options were specified")

    if options.regenerate_dhparams and not options.strong_dh:
        raise ValueError("--regenerate-dhparams should be specified only with --strong-dh")

    if (options.ca_chain or options.private_key) and not options.cert:
        raise ValueError("--cert should be specified")

    log.debug("ciphers: %s", ciphers)
    log.debug("protocols: %s", ', '.join(protocols) if protocols else 'None')

    if options.strong_dh:
        generate_dh_params(options.dhparams_size, options.regenerate_dhparams)

    rc = 0
    for service in services:
        if service not in modules.SERVICES.keys():
            log.warning("Ignore unknown service %s", service)
            continue

        module_name = modules.SERVICES.get(service)
        obj = module_name()

        if not obj.is_installed():
            log.debug("- %s\t\tnot installed", service)
            continue

        try:
            if options.show_custom:
                show_customized(service)
                continue

            if ciphers and protocols:
                log.debug("* %s [Configure Ciphers and Protocols]", service)

                if options.no_custom:
                    flush_custom_options(service)

                """ Working with cached ciphers and protocols """
                _ciphers, _protocols = get_custom_options(service, ciphers, protocols, options.custom)

                """ do action on service """
                obj.ciphers(_ciphers)
                obj.protocols(_protocols)

            if options.disable_tls_compression:
                log.debug("* %s [Disable TSL compression]", service)
                obj.tls_compression()

            if options.strong_dh:
                log.debug("* %s [Strong DH parameters]", service)
                obj.dh(options.dhparams_size)

            if options.certificate:
                log.debug("* %s [Install SSL certificate]", service)
                obj.certificate(options.cert, options.private_key, options.ca_chain)

            for name in obj.services_list():
                condrestart(name)

        except Exception, ex:
            log.error("failed to configure %s service: %s", service, ex)
            log.debug(ex, exc_info=True)
            rc = 1

    return rc

def main():
    """ entry point """
    try:
        exit_code = do_main()
    except Exception, ex:
        log.error('%s', ex)
        log.debug(ex, exc_info=True)
        exit_code = 1

    sys.exit(exit_code)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
